/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema.
 */

package org.objectweb.dream.message;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.dream.util.MultipleIterator;

/**
 * Implementation of extensible message with named sub messages.
 */
public class BasicExtensibleMessageNC extends BasicExtensibleMessage
    implements
      ExtensibleMessageNC,
      MessageTypeNC
{

  Map           subMessageMap           = new HashMap();
  Message[]     allSubMessagesArray     = null;
  MessageType[] allSubMessageTypesArray = null;

  // ---------------------------------------------------------------------------
  // Constructors
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   * 
   * @param messageManagerId the id of the message manager that creates this
   *          message.
   */
  public BasicExtensibleMessageNC(short messageManagerId)
  {
    super(messageManagerId);
  }

  // ---------------------------------------------------------------------------
  // Implementation of ExtensibleMessageNC interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.message.MessageNC#getSubMessage(java.lang.String)
   */
  public Message getSubMessage(String name)
  {
    return (Message) subMessageMap.get(name);
  }

  /**
   * @see org.objectweb.dream.message.MessageTypeNC#getSubMessageNames()
   */
  public String[] getSubMessageNames()
  {
    return (String[]) subMessageMap.keySet().toArray(EMPTY_STRING_ARRAY);
  }

  /**
   * @see org.objectweb.dream.message.MessageTypeNC#getSubMessageNamesIterator()
   */
  public Iterator getSubMessageNamesIterator()
  {
    return subMessageMap.keySet().iterator();
  }

  /**
   * @see org.objectweb.dream.message.ExtensibleMessageNC#addSubMessage(java.lang.String,
   *      org.objectweb.dream.message.Message)
   */
  public void addSubMessage(String name, Message message)
      throws MessageAlreadyExistException
  {
    subMessageMap.put(name, message);
    subMessagesArray = null;
    subMessageTypesArray = null;
    allSubMessagesArray = null;
    allSubMessageTypesArray = null;
  }

  /**
   * @see org.objectweb.dream.message.ExtensibleMessageNC#removeSubMessage(java.lang.String)
   */
  public Message removeSubMessage(String name)
  {
    Message removed = (Message) subMessageMap.remove(name);
    if (removed != null)
    {
      subMessagesArray = null;
      subMessageTypesArray = null;
      allSubMessagesArray = null;
      allSubMessageTypesArray = null;
    }
    return removed;
  }

  /**
   * @see org.objectweb.dream.message.MessageNC#getUnnamedSubMessageIterator()
   */
  public Iterator getUnnamedSubMessageIterator()
  {
    return super.getSubMessageIterator();
  }

  /**
   * @see org.objectweb.dream.message.MessageNC#getUnnamedSubMessages()
   */
  public Message[] getUnnamedSubMessages()
  {
    return super.getSubMessages();
  }

  // ---------------------------------------------------------------------------
  // Implementation of MessageTypeNC interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.message.MessageTypeNC#getSubMessageType(java.lang.String)
   */
  public MessageType getSubMessageType(String name)
  {
    Message msg = getSubMessage(name);
    if (msg == null)
    {
      return null;
    }
    return msg.getMessageType();
  }

  /**
   * @see org.objectweb.dream.message.MessageTypeNC#getUnnamedSubMessageTypes()
   */
  public MessageType[] getUnnamedSubMessageTypes()
  {
    return super.getSubMessageTypes();
  }

  /**
   * @see org.objectweb.dream.message.MessageTypeNC#getUnnamedSubMessageTypesIterator()
   */
  public Iterator getUnnamedSubMessageTypesIterator()
  {
    return super.getSubMessageTypesIterator();
  }

  // ---------------------------------------------------------------------------
  // Overriden methods
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.message.Message#getSubMessageIterator()
   */
  public Iterator getSubMessageIterator()
  {
    return new MultipleIterator(new Iterator[]{
        subMessageMap.values().iterator(), super.getSubMessageIterator()});
  }

  /**
   * @see org.objectweb.dream.message.Message#getSubMessages()
   */
  public Message[] getSubMessages()
  {
    if (allSubMessagesArray == null)
    {
      if (subMessages.isEmpty() && subMessageMap.isEmpty())
      {
        allSubMessagesArray = EMPTY_MESSAGE_ARRAY;
      }
      else
      {
        allSubMessagesArray = new Message[subMessages.size()
            + subMessageMap.size()];
        subMessages.toArray(allSubMessagesArray);
        Iterator iter = subMessageMap.values().iterator();
        int i = subMessages.size();
        while (iter.hasNext())
        {
          allSubMessagesArray[i] = (Message) iter.next();
          i++;
        }
      }
    }
    return allSubMessagesArray;
  }

  /**
   * @see org.objectweb.dream.message.MessageType#getSubMessageTypes()
   */
  public MessageType[] getSubMessageTypes()
  {
    if (allSubMessageTypesArray == null)
    {
      if (subMessages.isEmpty() && subMessageMap.isEmpty())
      {
        allSubMessageTypesArray = EMPTY_MESSAGE_TYPE_ARRAY;
      }
      else
      {
        allSubMessageTypesArray = new MessageType[subMessages.size()];
        Iterator iter = getSubMessageIterator();
        int i = 0;
        while (iter.hasNext())
        {
          Message message = (Message) iter.next();
          allSubMessageTypesArray[i] = message.getMessageType();
          i++;
        }
      }
    }
    return allSubMessageTypesArray;
  }

}