/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.dream.message;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import org.objectweb.dream.util.EmptyIterator;
import org.objectweb.dream.util.EmptyStringArray;

/**
 * Basic Implementation of {@link org.objectweb.dream.message.MessageType}
 */
public class BasicMessageType implements MessageType, EmptyStringArray
{

  /** A constant designating the empty message type. */
  public static final MessageType EMPTY_MESSAGE_TYPE   = new BasicMessageType();

  /** Hashtable associating names and chunk types */
  protected Hashtable             chunkTypes           = null;
  /** List of sub message types */
  protected LinkedList            subMessageTypes      = null;

  /** The chunk names */
  protected String[]              chkNames             = null;
  /** The sub message types */
  protected MessageType[]         subMessageTypesArray = null;

  /**
   * Creates an empty message type. This constructor is protected. To manipulate
   * the empty message type, components should use the
   * {@link #EMPTY_MESSAGE_TYPE}constant.
   */
  protected BasicMessageType()
  {
  }

  /**
   * Creates a message type with the specified chunk types and sub message
   * types. <code>chkNames</code> array contains the name of chunk types,
   * which are specified in <code>chkTypes</code> array (as a consequence
   * these two arrays must have the same length). Names and types are associated
   * following their order in arrays.
   * 
   * @param chkNames array of chunk names can be <code>null</code>.
   * @param chkTypes array of chunk types (ignored if <code>chkNames</code> is
   *          <code>null</code>).
   * @param msgTypes array of sub message types can be <code>null</code>.
   * @throws IllegalArgumentException If <code>chkNames</code> and
   *           <code>chkTypes</code> do not have the same length.
   * @throws ChunkAlreadyExistException If the same name appears twice in
   *           <code>chkNames</code>
   */
  public BasicMessageType(String[] chkNames, ChunkType[] chkTypes,
      MessageType[] msgTypes) throws ChunkAlreadyExistException
  {
    ChunkAlreadyExistException chunkAlreadyExistException = null;
    if (chkNames != null)
    {
      if (chkNames.length != chkTypes.length)
      {
        throw new IllegalArgumentException(
            "chkNames and chkTypes do not have the same length");
      }

      for (int i = 0; i < chkNames.length; i++)
      {
        try
        {
          addChunkType(chkNames[i], chkTypes[i]);
        }
        catch (ChunkAlreadyExistException e)
        {
          chunkAlreadyExistException = e;
        }
      }
    }

    if (msgTypes != null)
    {
      for (int i = 0; i < msgTypes.length; i++)
      {
        addSubMessageType(msgTypes[i]);
      }
    }
    if (chunkAlreadyExistException != null)
    {
      throw chunkAlreadyExistException;
    }
  }

  /**
   * Creates a message type that contains a chunk type. This constructor is
   * provided for convenience
   * 
   * @param chkName the name of the chunk type
   * @param chkType the chunk type
   */
  public BasicMessageType(String chkName, ChunkType chkType)
  {
    chunkTypes = new Hashtable();
    chunkTypes.put(chkName, chkType);
  }

  /**
   * Creates a message type that contains two chunk types. This constructor is
   * provided for convenience
   * 
   * @param chkName1 the name of the first chunk type
   * @param chkType1 the first chunk type
   * @param chkName2 the name of the second chunk type
   * @param chkType2 the second chunk type
   * @throws ChunkAlreadyExistException if <code>chkname1</code> and
   *           <code>chkName2</code> are equals
   */
  public BasicMessageType(String chkName1, ChunkType chkType1, String chkName2,
      ChunkType chkType2) throws ChunkAlreadyExistException
  {
    addChunkType(chkName1, chkType1);
    addChunkType(chkName2, chkType2);
  }

  /**
   * @see org.objectweb.dream.message.MessageType#getChunkType(java.lang.String)
   */
  public ChunkType getChunkType(String name)
  {
    if (chunkTypes == null)
    {
      return null;
    }
    return (ChunkType) chunkTypes.get(name);

  }

  /**
   * @see org.objectweb.dream.message.MessageType#getChunkNames()
   */
  public String[] getChunkNames()
  {
    if (chkNames == null)
    {
      if (chunkTypes == null)
      {
        chkNames = EMPTY_STRING_ARRAY;
      }
      else
      {
        chkNames = (String[]) chunkTypes.keySet().toArray(EMPTY_STRING_ARRAY);
      }
    }
    return chkNames;
  }

  /**
   * @see org.objectweb.dream.message.MessageType#getChunkNamesIterator()
   */
  public Iterator getChunkNamesIterator()
  {
    if (chunkTypes == null)
    {
      return EmptyIterator.INSTANCE;
    }
    return chunkTypes.keySet().iterator();
  }

  /**
   * @see org.objectweb.dream.message.MessageType#getSubMessageTypes()
   */
  public MessageType[] getSubMessageTypes()
  {
    if (subMessageTypesArray == null)
    {
      if (subMessageTypes == null)
      {
        subMessageTypesArray = EMPTY_MESSAGE_TYPE_ARRAY;
      }
      else
      {
        subMessageTypesArray = (MessageType[]) subMessageTypes
            .toArray(EMPTY_MESSAGE_TYPE_ARRAY);
      }
    }
    return subMessageTypesArray;
  }

  /**
   * @see org.objectweb.dream.message.MessageType#getSubMessageTypesIterator()
   */
  public Iterator getSubMessageTypesIterator()
  {
    if (subMessageTypes == null)
    {
      return EmptyIterator.INSTANCE;
    }
    return subMessageTypes.iterator();
  }

  /**
   * @see org.objectweb.dream.message.MessageType#isEmpty()
   */
  public boolean isEmpty()
  {
    return chunkTypes == null && subMessageTypes == null;
  }

  /**
   * @see org.objectweb.dream.message.MessageType#isSubTypeOf(org.objectweb.dream.message.MessageType)
   */
  public boolean isSubTypeOf(MessageType t)
  {
    if (t.isEmpty())
    {
      return true;
    }

    // test if every chunk types in t are in this type, and are sub types
    String[] tChkNames = t.getChunkNames();
    for (int i = 0; i < tChkNames.length; i++)
    {
      String name = tChkNames[i];
      ChunkType thisChkType = (ChunkType) chunkTypes.get(name);
      if (thisChkType == null)
      {
        // type t contains a chunk with a name that is not known as a chunk
        // name of this type.
        return false;
      }
      ChunkType tChkType = t.getChunkType(name);
      if (tChkType == null)
      {
        return false;
      }
      if (!thisChkType.isSubTypeOf(tChkType))
      {
        return false;
      }
    }

    // test if every message types in t are in this type, and are sub types
    // NOT IMPLEMENTED for sub message types
    return true;
  }

  /**
   * Adds a chunk type
   * 
   * @param chunkName the name of the chunk type
   * @param chunkType the chunk type
   * @throws ChunkAlreadyExistException if this type already contains a chunk
   *           type with the specified name
   */
  public void addChunkType(String chunkName, ChunkType chunkType)
      throws ChunkAlreadyExistException
  {
    if (chunkTypes == null)
    {
      chunkTypes = new Hashtable();
    }
    Object previousChunk = chunkTypes.put(chunkName, chunkType);
    if (previousChunk != null)
    {
      chunkTypes.put(chunkName, chunkType);
      throw new ChunkAlreadyExistException(
          "Chunk type with the same name already exist");
    }
    // invalidate chkNames array
    chkNames = null;
  }

  /**
   * Adds a sub message type
   * 
   * @param msgType the message type to add
   */
  public void addSubMessageType(MessageType msgType)
  {
    if (subMessageTypes == null)
    {
      subMessageTypes = new LinkedList();
    }
    subMessageTypes.add(msgType);
    // invalidate subMessageTypeArray
    subMessageTypesArray = null;
  }

  /**
   * removes a chunk type
   * 
   * @param name the name of the chunk type.
   * @return the chunk type or <code>null</code> if no chunk type with the
   *         specified name exists.
   */
  public ChunkType removeChunkType(String name)
  {
    if (chunkTypes == null)
    {
      return null;
    }
    ChunkType t = (ChunkType) chunkTypes.remove(name);
    if (t != null)
    {
      // invalidate chkNames array
      chkNames = null;
      if (chunkTypes.isEmpty())
      {
        chunkTypes = null;
      }
    }
    return t;
  }

  /**
   * Removes a sub message type
   * 
   * @param msgType the message type to remove
   * @return <code>true</code> if the specified message type can be removed
   */
  public boolean removeSubMessageType(MessageType msgType)
  {
    if (subMessageTypes == null)
    {
      return false;
    }
    if (subMessageTypes.remove(msgType))
    {
      // invalidate subMessageTypeArray
      subMessageTypesArray = null;
      if (subMessageTypes.isEmpty())
      {
        subMessageTypes = null;
      }
      return true;
    }
    return false;
  }
}